@@ -92,6 +92,7 @@ static char *logfile;
 
 #define print_output(...) fprintf (stdout, __VA_ARGS__), fflush(stdout)
 #define print_error(...) fprintf (stderr, __VA_ARGS__), fflush(stderr)
+#define absolute(_L, _i) (_i < 0 ? lua_gettop (_L) + _i + 1 : _i)
 
 static int colorize = 1;
 static char *colors[] = {"\033[0m",
@@ -106,7 +107,11 @@ static char *colors[] = {"\033[0m",
 
 #define COLOR(i) (colorize ? colors[i] : "")
 	
+#if LUA_VERSION_NUM == 501
+#define EOF_MARKER "'<eof>'"
+#else
 #define EOF_MARKER "<eof>"
+#endif
 
 static lua_State *_L;
 static int initialized = 0;
@@ -715,10 +720,7 @@ static void describe (lua_State *L, int index)
     size_t n;
     int type;
 
-    if (index < 0) {
-    	index = lua_gettop (L) + index + 1;
-    }
-    
+    index = absolute (L, index);    
     type = lua_type (L, index);
     
     if (type == LUA_TNUMBER) {
@@ -840,8 +842,16 @@ static void describe (lua_State *L, int index)
 		      COLOR(7), COLOR(8), lua_topointer (L, index));
 	dump_string (s, n);
     } else if (type == LUA_TUSERDATA) {
-	n = asprintf (&s, "<%suserdata:%s %p>",
-		      COLOR(7), COLOR(8), lua_topointer (L, index));
+	if (luaL_getmetafield (L, index, "__tostring")) {
+	    lua_pushvalue (L, index);
+	    lua_pcall (L, 1, 1, 0);
+	    s = (char *)lua_tolstring (L, -1, &n);
+	    lua_pop (L, 1);
+	} else {
+	    n = asprintf (&s, "<%suserdata:%s %p>",
+			  COLOR(7), COLOR(8), lua_topointer (L, index));
+	}
+	
 	dump_string (s, n);
     } else if (type == LUA_TTHREAD) {
 	n = asprintf (&s, "<%sthread:%s %p>",
@@ -994,6 +1004,8 @@ char *luap_describe (lua_State *L, int index)
 {
     struct winsize w;
 
+    index = absolute (L, index);    
+
     /* Initialize the state. */
     
     if (ioctl(STDOUT_FILENO, TIOCGWINSZ, &w) < 0) {
